#pragma once
#include <deque>
#include <string>
#include <sstream>

#include "math.h"

#include <complex>
using namespace std;

// special string type for case insensitive comparison
struct ci_char_traits : public std::char_traits<char>
                // just inherit all the other functions
                //  that we don't need to override
{
    static bool eq( char c1, char c2 )
        { return toupper(c1) == toupper(c2); }

    static bool ne( char c1, char c2 )
        { return toupper(c1) != toupper(c2); }

    static bool lt( char c1, char c2 )
        { return toupper(c1) <  toupper(c2); }

    static int compare( const char* s1,
                        const char* s2,
                        size_t n ) {
        return _memicmp( s1, s2, n );
                // if available on your compiler,
                //  otherwise you can roll your own
    }

    static const char*
    find( const char* s, int n, char a ) {
        while( n-- > 0 && toupper(*s) != toupper(a) ) {
            ++s;
        }
    return s;
    }
};

typedef std::basic_string<char, ci_char_traits> STArString;


// define the 1D and 2D smart arrays for storing values
// deque is used instead of usual containers like vector
// for more effeciency. check www.gotw.ca

template <typename T>
class Array1D
{
    public:
        static bool TestIndexInRange(const size_t index)
        {
            // size_t is assumed to be unsigned int (32 bits)
            // caution - for 64 bit OS, this will not work
            // we are checking the signed bit here
            // we may need to test the range as well in the future for
            // large values

            size_t test = index;
            test = (test >> 31);

            // if -ve index was passed in, then (index & 0x00000001) will be equal to 1 or true
            // so, in-range index should have this expression = false

            return !(test & 0x00000001);
        }

        Array1D() {}

        Array1D(int size) { resize(size); }

        typedef typename std::deque<T>::size_type size_type;

        const T& operator[] (const size_t i) const
        {
            size_type index = TestIndexInRange(i) ? i : 0;

            // if the array size is 0
            if(array_.size() == 0)
            {
                // we have to make at least the size = 1
               const_cast<std::deque<T> &>(array_).resize(1);
            }

            if(array_.size() < (index + 1))
                index = array_.size() - 1;

            return array_.at(index);
        }

        T& operator[](const size_t i)
        {
            size_type index = TestIndexInRange(i) ? i : 0;

            // automatically resize the array
            // on demand
            if(array_.size() < (index + 1))
                array_.resize(index + 1);

            return array_.at(index);
        }


        // the size of the array
        size_type size() const
        {
            return array_.size();
        }

        // call this method to initialize the size of the array
        // it is more effecient to set the size of the array before
        // doing operations
        void resize(const size_t new_size)
        {
            size_type newsize = TestIndexInRange(new_size) ? new_size : 0;

            array_.resize(newsize);
        }

        // assignment operator
        Array1D<T>& operator=(const Array1D<T>& source)
        {
            // clear the current data completely
            std::deque<T>().swap(array_);

            // assign the source data
            std::copy(source.array_.begin(), source.array_.end(), std::back_inserter(array_));

            return *this;
        }

        // copy constructor
        // especially important for 2D array implementation
        Array1D(const Array1D<T>& source)
        {
            // for copy constructor, the array_ is empty
            // so, no need to clear

            // assign the source data
            std::copy(source.array_.begin(), source.array_.end(), std::back_inserter(array_));
        }

        // swap function - for fast transferring data
        // from one object to another
        // caution - not like copy or =
        // it will destroy information on the 'source' object
        void swap(Array1D<T> &source)
        {
            array_.swap(source.array_);
        }

        // convert the array to a string
        std::string toString() const
        {
			std::ostringstream oss;

            size_t length = array_.size();
            for(size_t index = 0; index < length; ++index)
            {
                oss << array_[index];

				if(index < (length - 1))
					oss << ",";
            }
            return oss.str();
        }

        template<typename U>
    	void Convert1DFromString(const std::string &str, U& val)
    	{
    		std::istringstream iss(str);
    		iss >> val;
		}

		template<>
		void Convert1DFromString(const std::string &str, int& val)
    	{
			char *pstr_stop = 0;
			val = static_cast<int>(strtod(str.c_str(), &pstr_stop));
		}

		template<>
		void Convert1DFromString(const std::string &str, double& val)
    	{
			//char *pstr_stop = 0;
			//val = strtod(str.c_str(), &pstr_stop);
            char  sign = 0;
            const char *cp = str.c_str();
            
            if (*cp == '+' || *cp == '-')
                sign = *cp++;

            if (_stricmp(cp, "1.#inf") == 0) {
                if (!sign || sign == '+')
                {
                    unsigned long long raw = 0x7ff0000000000000;
                    val = *( double* )&raw; /* assign +Inf */
                }
                else
                {
                    unsigned long long raw = 0xfff0000000000000;
                    val = *( double* )&raw; /* assign -Inf */
                }
            }
            else if (_stricmp(cp, "nan") == 0)
            {
                unsigned long long raw = 0x7fc00000;
                val = *( double* )&raw; /* assign NaN */
            }
            else /* valid number */
            {
                char *pstr_stop = 0;
                val = strtod(str.c_str(), &pstr_stop);
            }

		}


        template<>
    	void Convert1DFromString(const std::string &str, STArString& val)
    	{
    		val.clear();
    		for ( std::string::const_iterator str_Iter = str.begin( ); str_Iter != str.end( ); str_Iter++ )
    		{
    			val.push_back(*str_Iter);
    		}
    	}

        template<>
    	void Convert1DFromString(const std::string &str, std::string& val)
    	{
    		val = str;
    	}
    
        // populate from string
        void fromString(const std::string &str)
        {
			// reset the contents
            std::deque<T>().swap(array_);

			std::istringstream iss;
            iss.clear();

			std::string separators(",");
            size_t n = str.length();

            size_t start = 0;
            size_t stop  = 0;
            
			T val;

			while ((start != std::string::npos) && (start < n))
            {
                stop = str.find_first_of(separators, start);
                if ((stop < 0) || (stop > n)) stop = n;

				Convert1DFromString(str.substr(start, stop - start), val);

                array_.push_back(val);

                start = stop + 1;
            }
        }

        // have to make the array_ public data type
        // so that codec can be effecient
        std::deque<T> array_;

};

template <> 
void Array1D < std::complex<double> >::fromString(const std::string &str)
{
	//TRACE(_T("*************************Complex Array1D fromString*************************"));

	// reset the contents
	std::deque<std::complex<double>>().swap(array_);

	std::istringstream iss;
	iss.clear();

	std::string hash("#");
	size_t realStart	= 0;
	size_t realEnd		= 0;
	size_t n			= str.length();

	realEnd = str.find_first_of(hash, realStart);
	std::string strReal("");
	std::string strImaginary("");

	strReal = str.substr(realStart, realEnd);
	strImaginary = str.substr(realEnd + 1);

	//size_t nReal = strReal.length();
	//size_t nImaginary = strImaginary.length();

	std::string separators(",");

	size_t startR = 0;
	size_t stopR = 0;

	size_t startI = 0;
	size_t stopI = 0;

	double valR = 0;
	double valI = 0;

	while (((startR != std::string::npos) && (startR < realEnd)) || ((startI != std::string::npos) && (startI < n)))
	{
		stopR = strReal.find_first_of(separators, startR);
		if ((stopR < 0) || (stopR > realEnd)) stopR = realEnd;

		stopI = strImaginary.find_first_of(separators, startI);
		if ((stopI < 0) || (stopI > n)) stopI = n;

		Convert1DFromString(strReal.substr(startR, stopR - startR), valR);
		Convert1DFromString(strImaginary.substr(startI, stopI - startI), valI);

		std::complex<double> val(valR,valI);

		array_.push_back(val);

		startR = stopR + 1;
		startI = stopI + 1;
	}
}

template <>
std::string Array1D < std::complex<double> >::toString() const
{
	//TRACE(_T("*************************Complex Array1D toString*************************"));

	std::ostringstream oss;
	std::ostringstream ossR;
	std::ostringstream ossI;

	size_t length = array_.size();
	for (size_t index = 1; index <= length; ++index)
	{
		ossR << array_[index - 1].real();
		ossI << array_[index - 1].imag();

		//if (index < (length - 1)) {
			ossR << ",";
			ossI << ",";
		//}
	}
	
	std::string str_real(ossR.str());
	std::string str_imag(ossI.str());

	oss << str_real.substr(0, str_real.length() - 1);
	oss << "#";
	oss << str_imag.substr(0, str_imag.length() - 1);

	return oss.str();
}

template <typename T>
class Array2D
{
    public:
        Array2D():data_(0) {}

        Array2D(int rows, int cols) { resize(rows, cols); }

        typedef typename std::deque<T>::size_type size_type;

        //-----------------------------------------------------------------
        //  The class's [] operator returns a row of the 2d array.
        //  std::deque's own [] operator will take care of returning
        //  an individual item of the 2d array when we use the [][] notation.
        //-----------------------------------------------------------------

        const Array1D<T>& operator[] ( const size_t i ) const
        {
            // for const array element access

            size_type index = Array1D<T>::TestIndexInRange(i) ? i : 0;

            // if the array size is 0
            if(data_.size() == 0)
            {
                // we have to make at least the size = 1
               const_cast<std::deque< Array1D<T> > &>(data_).resize(1);
            }

            if(data_.size() < (index + 1))
                index = data_.size() - 1;

            return data_.at(index);
        }

        Array1D<T>& operator[] ( const size_t i )
        {
            size_type index = Array1D<T>::TestIndexInRange(i) ? i : 0;

            // automatically resize the array
            // on demand
            if(data_.size() < (index + 1))
                data_.resize(index + 1);

            return data_.at(index);
        }

        // the size of the array
        size_type rows() const
        {
            return data_.size();
        }

        // number of columns for each row can be skewed
        size_type cols(const size_t row = 0) const
        {
            if( !Array1D<T>::TestIndexInRange(row) )
                return 0;

            if(data_.size() <= row) 
                return 0;

            return data_.at(row).size();
        }

        // call this method to initialize the size of the array
        // it is more effecient to set the size of the array before
        // doing operations

        void resize(const size_t newsize_r, const size_t newsize_c)
        {
            size_type newsize_row = Array1D<T>::TestIndexInRange(newsize_r) ? newsize_r : 0;
            size_type newsize_col = Array1D<T>::TestIndexInRange(newsize_c) ? newsize_c : 0;

            // step 1: resize the current columns for each current row
            size_type current_rows = data_.size();
            for(size_t index = 0; index < current_rows; ++index)
            {
                data_.at(index).resize(newsize_col);
            }

            // step 2: resize the rows to the requested size
            data_.resize(newsize_row);

            // step 3: if new rows are added, resize the columns in the rows
            for(size_t index = current_rows; index < newsize_row; ++index)
            {
                data_.at(index).resize(newsize_col);
            }
        }

        // resize any columns for a row
        void resize_column(const size_t row, const size_t newsize_col)
        {
            if( !Array1D<T>::TestIndexInRange(row) )
                return;

            if( data_.size() <= row ) return;

            data_.at(row).resize(newsize_col);
        }

        // resize the number of rows
        void resize_row(const size_t newsize_r)
        {
            size_type newsize_row = Array1D<T>::TestIndexInRange(newsize_r) ? newsize_r : 0;
            data_.resize(newsize_row);
        }

        // assignment operator
        Array2D<T>& operator=(const Array2D<T>& source)
        {
            // clear the current data completely
            std::deque< Array1D<T> >().swap(data_);

            // assign the source data
            std::copy(source.data_.begin(), source.data_.end(), std::back_inserter(data_));

            return *this;
        }

        // copy constructor
        // especially important for 2D array implementation
        Array2D(const Array2D<T>& source)
        {
            // for copy constructor, the data_ is empty
            // so, no need to clear

            // assign the source data
            std::copy(source.data_.begin(), source.data_.end(), std::back_inserter(data_));
        }

        // swap function - for fast transferring data
        // from one object to another
        // caution - not like copy or =
        // it will destroy information on the 'source' object
        void swap(Array2D<T> &source)
        {
            data_.swap(source.data_);
        }

        // convert the array to a string
        std::string toString() const
        {
			std::ostringstream oss;

            size_type current_rows = data_.size();
            for(size_t index = 0; index < current_rows; ++index)
            {
                oss << data_.at(index).toString();
                
                if(index < (current_rows-1))
                    oss << "|";
            }

            return oss.str();
        }

        // populate from string
        void fromString(const std::string &str)
        {

			// reset the contents
            std::deque< Array1D<T> >().swap(data_);

            // we need to scan the string manually

            std::string separators("|");
            size_t n = str.length();

            size_t start = 0;
            size_t stop  = 0;
            
			while ((start != std::string::npos) && (start < n))
            {
                stop = str.find_first_of(separators, start);
                if ((stop < 0) || (stop > n)) stop = n;

                Array1D<T> arr1d;

                arr1d.fromString(str.substr(start, stop - start));
                

                data_.push_back(arr1d);

                start = stop + 1;
            }
        }

		

        // have to make the data_ public data type
        // so that codec can be effecient
        std::deque< Array1D<T> > data_;

};

template <>
void Array2D < std::complex<double> >::fromString(const std::string &str)
{
	//TRACE(_T("*************************Complex Array1D fromString*************************"));

	// reset the contents

	data_.clear();

	std::basic_string<char>::size_type pos = str.find("#");
	if (pos != std::basic_string<char>::npos)
	{
		std::string real_str = str.substr(0, pos);
		std::string imag_str = str.substr(pos + 1);

		real_str += "|";
		imag_str += "|";

		std::basic_string<char>::size_type start_row = 0;
		std::basic_string<char>::size_type end_row = real_str.find("|");

		if (end_row == std::basic_string<char>::npos && !real_str.empty())
		{
			data_[0][0].real(atof(real_str.c_str()));
		}
		else
		{
			size_t rcount = 0;
			while (end_row != std::string::npos)
			{
				std::string real_row((real_str.substr(start_row, (end_row - start_row))));

				real_row += ",";

				std::basic_string<char>::size_type start_col = 0;
				std::basic_string<char>::size_type end_col = real_row.find(",");

				

				Array1D<std::complex<double> > arr1d;
				if (data_.size() <= rcount)
				{
					data_.push_back(arr1d);
				}

				size_t count = 0;
				while (end_col != std::string::npos)
				{
					if (end_col > start_col)
					{
						std::string real_data((real_row.substr(start_col, (end_col - start_col))));

						Array1D<std::complex<double> > &arr = data_[rcount];

						arr[count].real(atof(real_data.c_str()));
					}

					count++;
					start_col = end_col + 1;
					end_col = real_row.find(",", start_col);
				}

				rcount++;
				start_row = end_row + 1;
				end_row = real_str.find("|", start_row);
			}
		}

		start_row = 0;
		end_row = imag_str.find("|");
		if (end_row == std::basic_string<char>::npos && !imag_str.empty())
		{
			data_[0][0].imag(atof(imag_str.c_str()));
		}
		else
		{
			size_t rcount = 0;
			while (end_row != std::string::npos)
			{
				std::string imag_row((imag_str.substr(start_row, (end_row - start_row))));

				imag_row += ",";

				std::basic_string<char>::size_type start_col = 0;
				std::basic_string<char>::size_type end_col = imag_row.find(",");

				Array1D<std::complex<double> > arr1d;
				if (data_.size() <= rcount)
				{
					data_.push_back(arr1d);
				}

				size_t count = 0;
				while (end_col != std::string::npos)
				{
					if (end_col > start_col)
					{
						std::string imag_data((imag_row.substr(start_col, (end_col - start_col))));
						Array1D<std::complex<double> > &arr = data_[rcount];
						arr[count].imag(atof(imag_data.c_str()));
					}
					count++;
					start_col = end_col + 1;
					end_col = imag_row.find(",", start_col);
				}

				rcount++;
				start_row = end_row + 1;
				end_row = imag_str.find("|", start_row);
			}
		}
	}
}

template <>
std::string Array2D < std::complex<double> >::toString() const
{
	//TRACE(_T("*************************Complex Array1D toString*************************"));

	std::ostringstream oss;
	std::ostringstream ossR;
	std::ostringstream ossI;

	size_t rows = data_.size();

	for (size_t i = 0; i < rows; i++)
	{
		std::ostringstream ossRR;
		std::ostringstream ossII;

		Array1D<std::complex<double> > arr = data_.at(i);
		size_t cols = arr.size();
		for (size_t j = 0; j < cols; j++)
		{
			ossRR << arr[j].real() << ",";
			ossII << arr[j].imag() << ",";
		}


		ossR << ossRR.str().substr(0, ossRR.str().length() - 1) << "|";
		ossI << ossII.str().substr(0, ossII.str().length() - 1) << "|";
	}


	std::string str_real(ossR.str());
	std::string str_imag(ossI.str());

	oss << str_real.substr(0, str_real.length() - 1);
	oss << "#";
	oss << str_imag.substr(0, str_imag.length() - 1);

	return oss.str();
}

typedef Array1D<int>			 Array1D_Integer;
typedef Array1D<long>			 Array1D_Long;
typedef Array1D<double>			 Array1D_Double;
typedef Array1D<std::complex<double>> Array1D_Complex;
typedef Array1D<std::string>	 Array1D_String;
typedef Array1D<bool>			 Array1D_Bool;

typedef Array2D<int>			 Array2D_Integer;
typedef Array2D<long>			 Array2D_Long;
typedef Array2D<double>			 Array2D_Double;
typedef Array2D<std::complex<double>> Array2D_Complex;
typedef Array2D<std::string>	 Array2D_String;
typedef Array2D<bool>			 Array2D_Bool;